from urllib.parse import urljoin
from dataclasses import dataclass
import requests
import sseclient
import json
import threading
import time


@dataclass
class Metadata:
    name: str
    type: str
    readonly: bool


class ServerSentEventStream:

    def __init__(self, openhab_uri: str, itemname: str, on_item_changed_callback):
        self.itemname = itemname
        self.on_item_changed_callback = on_item_changed_callback
        self.openhab_item_event_uri = urljoin(openhab_uri, '/rest/events?topics=smarthome/items/' + itemname + '/state')
        self.is_running = True
        self.thread = None

    def start(self):
        self.thread = threading.Thread(target=self.__listen)
        self.thread.start()

    def __listen(self):
        while self.is_running:
            try:
                response = requests.get(self.openhab_item_event_uri, stream=True)
                client = sseclient.SSEClient(response)
                try:
                    for event in client.events():
                        data = json.loads(event.data)
                        payload = json.loads(data['payload'])
                        value = payload['value']
                        print("openhab item " + self.itemname + " has been updated to " + str(value))
                        self.on_item_changed_callback(value)
                finally:
                    client.close()
            except Exception as e:
                print("error occurred consuming sse for " + self.itemname + " " + str(e))
                time.sleep(5)

    def stop(self):
        self.is_running = False
        threading.Thread.join(self.thread)


class OpenhabItem:

    def __init__(self, openhab_uri: str, itemname: str):
        self.openhab_uri = openhab_uri
        self.itemname = itemname
        self.openhab_item_uri = urljoin(self.openhab_uri, '/rest/items/' + self.itemname)
        self.__metadata = None

    def metadata(self) -> Metadata:
        if self.__metadata is None:
            print('openhab item ' + self.name + ' fetching meta data')
            resp = requests.get(self.openhab_item_uri)
            resp.raise_for_status()
            data = json.loads(resp.text)
            type = data['type'].lower()
            readonly = False
            self.__metadata = Metadata(self.name, type, readonly)
            print('openhab item ' + self.name + " meta data loaded (type: " + self.__metadata.type + ", readonly: " + str(self.__metadata.readonly) + ")")
        return self.__metadata

    @property
    def type(self) -> str:
        return self.metadata().type

    @property
    def writeable(self) -> bool:
        return not self.metadata().readonly

    @property
    def name(self) -> str:
        return self.itemname

    @property
    def state(self):
        try:
            resp = requests.get(self.openhab_item_uri + '/state')
            resp.raise_for_status()
            value = resp.text
            print("openhab item " + self.itemname + " read " + str(value))
            return value
        except requests.exceptions.HTTPError as err:
            print("got error by reading openhab item " + self.itemname + " reason: " + resp.text)

    @state.setter
    def state(self, value):
        try:
            print("writing openhab item " + self.itemname + " with " + str(value))
            resp = requests.put(self.openhab_item_uri + '/state', data=str(value), headers={'Content-Type': 'text/plain'})
            resp.raise_for_status()
        except requests.exceptions.HTTPError as err:
            print("got error by writing openhab item " + self.itemname + " = " + str(value) + " reason: " + resp.text)

    def new_change_listener(self, on_changed_callback) -> ServerSentEventStream:
        return ServerSentEventStream(self.openhab_uri, self.itemname, on_changed_callback)





import argparse
from pushgateway.gateway import run
from pushgateway.config import default_config_file
from pushgateway.unit import register, deregister, printlog

PACKAGENAME = 'pushgateway'
ENTRY_POINT = "pushgateway"
DESCRIPTION = "A push gateway sending webthing properties updates to a openhab server"


def main():
    parser = argparse.ArgumentParser(description=DESCRIPTION)
    parser.add_argument('--command', metavar='command', required=True, type=str, help='the command. Supported commands are: listen (run the service), register (register and starts the service as a systemd unit, deregister (deregisters the systemd unit), log (prints the log)')
    parser.add_argument('--filename', metavar='filename', required=False, type=str,  help='the config filename')
    args = parser.parse_args()

    if args.filename is None:
        filename = default_config_file()
    else:
        filename = args.filename

    if args.command == 'listen':
        print("running " + PACKAGENAME + " with config " + filename)
        run(filename)
    elif args.command == 'register':
        print("register " + PACKAGENAME + " with config " + filename)
        register(PACKAGENAME, ENTRY_POINT, filename)
    elif args.command == 'deregister':
        deregister(PACKAGENAME)
    elif args.command == 'log':
        printlog(PACKAGENAME)
    else:
        print("usage " + ENTRY_POINT + " --help")


if __name__ == '__main__':
    main()




from dataclasses import dataclass
from typing import List
import os
import logging
import pathlib


@dataclass
class Config:
    webthing_root_uri: str
    webthing_property_name: str
    openhab_root_uri: str
    openhab_item_name: str


def default_config_file():
    filename = pathlib.Path(os.getcwd(), "gateway.conf")
    if not filename.exists():
        filename.parent.mkdir(parents=True, exist_ok=True)
        with open(filename, "w") as file:
            file.write("# webthing_root_uri, webthing_property_name, openhab_root_uri, openhab_item_name")
        logging.info("config file " + str(filename) + " generated")
    return str(filename)


def load_config(filename: str) -> List[Config]:
    config = list()
    with open(filename, "r") as file:
        for line in file.readlines():
            line = line.strip()
            if not line.startswith("#") and len(line) > 0:
                try:
                    parts = line.split(",")
                    config.append(Config(parts[0].strip(), parts[1].strip(), parts[2].strip(), parts[3].strip()))
                except Exception as e:
                    logging.error("invalid syntax in line " + line + "  ignoring it" + str(e))
    return config



from pushgateway.config import load_config
from pushgateway.webthing import WebthingProperty
from pushgateway.openhab import OpenhabItem
import time
import threading


class WebThingPropertyToOpenhabItemLink:

    def __init__(self, webthing_property: WebthingProperty, openhab_item: OpenhabItem):
        self.webthing_property = webthing_property
        self.openhab_item = openhab_item
        self.cached_value = None

    def start(self):
        threading.Thread(target=self.__listen).start()

    def __listen(self):
        while True:
            try:
                # listen for values to forwards
                stream = self.webthing_property.new_change_listener(self.__on_changed_callback)
                stream.start()

                # set initial state (if changes are coming sporadically)
                self.openhab_item.state = self.__convert(self.webthing_property.property)

                time.sleep(10 * 60)
                stream.stop()
            except Exception as e:
                print("error occurred for webthing " + self.webthing_property.name + ": "+ str(e))
                time.sleep(10)

    def __on_changed_callback(self, new_value):
        item_value = self.__convert(new_value)
        if self.cached_value != item_value:
            self.cached_value = item_value
            self.openhab_item.state = item_value

    def __convert(self, property_value):
        if self.webthing_property.type == 'boolean':
            if property_value:
                return "ON"
            else:
                return "OFF"
        else:
            return property_value



class OpenhabItemToWebThingPropertyLink:

    def __init__(self, webthing_property: WebthingProperty, openhab_item: OpenhabItem):
        self.webthing_property = webthing_property
        self.openhab_item = openhab_item
        self.cached_value = None

    def start(self):
        threading.Thread(target=self.__listen).start()

    def __listen(self):
        while True:
            try:
                # listen for values to forwards
                stream = self.openhab_item.new_change_listener(self.__on_changed_callback)
                stream.start()

                # set initial state (if changes are coming sporadically)
                self.webthing_property.property = self.__convert(self.openhab_item.state)

                time.sleep(10 * 60)
                stream.stop()
            except Exception as e:
                print("error occurred for openhab " + self.webthing_property.name + ": "+ str(e))
                time.sleep(10)

    def __on_changed_callback(self, new_value):
        property_value = self.__convert(new_value)
        if self.cached_value != property_value:
            self.cached_value = property_value
            self.webthing_property.property = property_value

    def __convert(self, value):
        source_type = self.openhab_item.type
        target_type = self.webthing_property.type
        if source_type == 'switch':
            return value == 'ON'
        elif target_type == 'number':
            return float(value)
        elif  target_type == 'integer':
            return int(value)
        else:
            return value


class Link:

    def __init__(self, webthing_uri: str, webthing_property: str, openhab_uri: str, itemname: str):
        self.webthing_property = WebthingProperty(webthing_uri, webthing_property)
        self.openhab_item = OpenhabItem(openhab_uri, itemname)

    def start(self):
        threading.Thread(target=self.__listen).start()

    def __listen(self):
        if self.webthing_property.writeable:
            openhab_to_webthing_link = OpenhabItemToWebThingPropertyLink(self.webthing_property, self.openhab_item)
            openhab_to_webthing_link.start()
        else:
            webthing_to_openhab_link = WebThingPropertyToOpenhabItemLink(self.webthing_property, self.openhab_item)
            webthing_to_openhab_link.start()

        while True:
            time.sleep(5)


def run(filename: str):
    for config in load_config(filename):
        Link(config.webthing_root_uri, config.webthing_property_name, config.openhab_root_uri, config.openhab_item_name).start()

    while True:
        time.sleep(60)



from urllib.parse import urljoin
from dataclasses import dataclass
import requests
import websocket
import json
import time
import threading


@dataclass
class Metadata:
    name: str
    type: str
    readonly: bool
    prop_uri: str
    prop_ws_uri: str


class WebSocketStream:

    def __init__(self, metadata: Metadata, on_property_changed_callback):
        self.metadata = metadata
        self.on_property_changed_callback = on_property_changed_callback
        self.is_running = True
        self.thread = None

    def start(self):
        self.thread = threading.Thread(target=self.__listen)
        self.thread.start()

    def __listen(self):
        while self.is_running:
            try:
                ws = websocket.WebSocket()
                try:
                    ws.connect(self.metadata.prop_ws_uri)
                    print('webthing property ' + self.metadata.name + ' websocket ' + self.metadata.prop_ws_uri + ' connected')
                    while self.is_running:
                        msg = json.loads(ws.recv())
                        if msg['messageType'] == 'propertyStatus':
                            data = msg['data']
                            if self.metadata.name in data.keys():
                                value = data[self.metadata.name]
                                print("webthing property " + self.metadata.name + " has been updated to " + str(value))
                                self.on_property_changed_callback(value)
                finally:
                    print('websocket ' + self.metadata.prop_ws_uri + ' disconnected')
                    ws.close()
            except Exception as e:
                print("error occurred consuming web socket for " + self.metadata.name + " " + str(e))
                time.sleep(5)

    def stop(self):
        self.is_running = False
        threading.Thread.join(self.thread)


class WebthingProperty:

    def __init__(self, webthing_uri: str, webthing_property: str):
        self.__webthing_uri = webthing_uri
        self.__webthing_property = webthing_property
        self.__metadata = None

    def metadata(self) -> Metadata:
        if self.__metadata is None:
            print('webthing property ' + self.__webthing_property + ' fetching meta data')
            webthing_meta = requests.get(self.__webthing_uri).json()
            props = webthing_meta['properties'][self.__webthing_property]
            webthing_type = props['type']
            if 'readOnly' in props.keys():
                webthing_readonly = props['readOnly']
            else:
                webthing_readonly = False
            webthing_prop_uri = None
            for link in props['links']:
                if 'rel' in link.keys():
                    if link['rel'] == 'property':
                        webthing_prop_uri = urljoin(self.__webthing_uri, link['href'])
            webthing_prop_ws_uri = None
            for link in webthing_meta['links']:
                if 'rel' in link.keys():
                    if link['rel'] == 'property':
                        webthing_prop_uri = urljoin(self.__webthing_uri, link['href'])
                    elif link['rel'] == 'alternate':
                        webthing_prop_ws_uri = urljoin(self.__webthing_uri, link['href'])
            self.__metadata = Metadata(self.__webthing_property, webthing_type, webthing_readonly, webthing_prop_uri, webthing_prop_ws_uri)
            print('webthing property ' + self.__webthing_property + " meta data loaded (type: " + self.__metadata.type + ", readonly: " + str(self.__metadata.readonly) + ")")
        return self.__metadata

    @property
    def name(self) -> str:
        return self.metadata().name

    @property
    def type(self) -> str:
        return self.metadata().type

    @property
    def writeable(self) -> bool:
        return not self.metadata().readonly

    @property
    def property(self):
        properties = requests.get(self.metadata().prop_uri).json()
        value =  properties[self.metadata().name]
        print("webthing property " + self.metadata().name + " read " + str(value))
        return value

    @property.setter
    def property(self, value):
        try:
            print("writing webthing property " + self.name + " with " + str(value))
            body = json.dumps({ self.metadata().name: value }, indent=2)
            resp = requests.put(self.metadata().prop_uri, data=body, headers={'Content-Type': 'application/json'})
            resp.raise_for_status()
        except requests.exceptions.HTTPError as err:
            print("got error by webthing property " + self.name + " = " + str(value) + " reason: " + resp.text)

    def new_change_listener(self, on_changed_callback) -> WebSocketStream:
        return WebSocketStream(self.metadata(), on_changed_callback)
